/*     ************************************************************
SYNOPSIS:  This program approximates the sampling probability of an ordered
           configuration using the asymptotic expansion discussed in
           Jenkins and Song (2009).
	   Either an infinite-alleles or a finite-alleles PIM may be used.

           asf [-k K] [-l L] [-a a_1 ... a_K] [-b b_1 ... b_L]
               [-P P_1 ... P_K] [-Q Q_1 ... Q_L] [-tA thetaA] [-tB thetaB]
               [-r rho1 rho2 ... ] [-h]
               -c c_{11}...c_{1L} c_{21}...c_{2L} ... c_{K1}...c_{KL}
                 

OPTIONS:  
   -k #             Number of alleles at locus A.  (default = 2)
                    This option is ignored if -a option is
                    specified.   Either -a or -k must be specified.

   -l #             Number of alleles at locus B.  (default = 2)
                    This option is ignored if -b option is
                    specified.  Either -b or -l must be specified.

   -a  a_1 ... a_K  Configuration for left-half fragments.

   -b  b_1 ... b_L  Configuration for right-half fragments.

   -c  c_11...c_KL  Configuration for full gametes.

   -P  P_1 ... P_K  Mutation transition matrix for locus A.
                    If either -P or -Q is specified, then PIM model is
                    assumed.

   -Q  Q_1 ... Q_K  Mutation transition matrix for locus B.

                    If either -P or -Q is specified, then PIM model is
                    assumed.

   -r  # ... #      Recombination rates (default = 50).

   -tA #            Mutation rate at locus A (default = 0.001).

   -tB #            Mutation rate at locus B (default = 0.001).

   -d  #            Display option.
                    Input a number greater than 0 for condensed output.

   -e               Exact computation of q_2.  If this option is not specified,
                    the recursive part of q_2 will not be computed.

   -h               Print this page.


-------------------------
Authors : Yun S. Song
          Paul A. Jenkins
          2009
************************************************************  */

#include <iostream>
#include <math.h>
#include <vector>
#include <sstream>
#include <stdio.h>
#include <time.h>
#include <fstream>
#include "basicdefs.h"
#include "getq.h"
#include "weakComposition.h"
#include "dynamicProgramming.h"
#include "integerPartitions.h"

using namespace std;
vector<int> aVec, bVec, cVec;
vector<vector<int> > cMat;
int noAllelesA=2,  noAllelesB=2;
double thetaA=0.001, thetaB=0.001;
vector<double> P, Q, rho;
int display = 0;
int aTotal = 0, bTotal = 0, cTotal = 0;
bool exactq2 = false; // whether to compute the recursive part of q_2

int main(int argc, char *argv[]){
  int iI, iJ;
  getoptions(argc,argv);
  if(rho.size() == 0) rho.push_back(50.0);
  if(P.size() > 0 && Q.size() > 0)
  {
    double sumP = 0.0, sumQ = 0.0;
    for (iI = 0; iI<P.size(); iI++) sumP += P[iI];
    for (iJ = 0; iJ<Q.size(); iJ++) sumQ += Q[iJ];
    
    if (sumP != 1.0){
      // Allows for numerical errors or non-normalized transition matrices
      // for (iI = 0; iI<P.size(); iI++) P[iI] /= sumP;	
      cerr << "P_i don't sum to 1." <<endl;
      exit(1);
    }

	  
    if (sumQ != 1.0){
      // Allows for numerical errors or non-normalized transition matrices
      // for (iJ = 0; iJ<Q.size(); iJ++) Q[iJ] /= sumQ;
      cerr << "Q_i don't sum to 1." <<endl;
      exit(1);
    }

  }
  else if (P.size() > 0 || Q.size() > 0)
  {
	cerr << "\nError: Need to input a transition matrix for each locus" << endl;
	exit(1);
  }
	
  double pzero, pone, ptwoEg, ptwoRec=0.0, asf;

  vector<int> max(4);
  vector<vector<int> > config;
  

  // check aVec and the number of alleles at locus A
  if(aVec.size() != 0){
    noAllelesA = aVec.size();
  }
  else{
    aVec.resize(noAllelesA);
    fill( aVec.begin(), aVec.end(), 0 );
  }
  
  // check bVec and the number of alleles at locus B
  if(bVec.size() != 0){
    noAllelesB = bVec.size();
  }
  else{
    bVec.resize(noAllelesB);
    fill( bVec.begin(), bVec.end(), 0 );
  }
  
  // create cMatrix
  if(noAllelesA*noAllelesB != cVec.size()){
    cerr << "Incorrect a,b,c configuration." << endl;
    exit(1);
  }
  else if ((P.size() > 0) && (noAllelesA != P.size() || noAllelesB != Q.size())){
    cerr << "\nError: Mutation transition matrices are the wrong size." << endl;
    exit(1);
  }
  else{
    cMat.resize(noAllelesA);
    for(iI=0; iI < noAllelesA; iI++){
      cMat[iI].resize(noAllelesB);
      for(iJ=0; iJ < noAllelesB; iJ++){
	  cMat[iI][iJ]=cVec[iI*noAllelesB+iJ];
      }
    }
  }
  
  pzero=qzero(aVec,bVec,cMat,thetaA,thetaB,P,Q);
  pone=qone(aVec,bVec,cMat,thetaA,thetaB,P,Q);
  ptwoEg=qtwoEg(aVec,bVec,cMat,thetaA,thetaB,P,Q);
  if(exactq2 == true){
    if(noAllelesB <= 2 && noAllelesA <=2){
      ptwoRec=qtwoRec2(aVec,bVec,cMat,thetaA,thetaB,P,Q);
    }
    else{
      ptwoRec=qtwoRec(aVec,bVec,cMat,thetaA,thetaB,P,Q);
    }
  }
  
  if(display == 0){
    // print input 
    cout << "aVector = "; printVec(aVec);  cout << endl;
    cout << "bVector = "; printVec(bVec);  cout << endl;
    cout << "cMatrix = "; printVec(cMat[0]);
    for(iI=1; iI < cMat.size(); iI++){
      cout << "          "; printVec(cMat[iI]);
    }
    printf("\nMultiplicity (a,b,c) = (%i, %i, %i)\n",aTotal,bTotal,cTotal);

    cout << endl;
    printf("thetaA = %.3e, thetaB = %.3e\n\n", thetaA,thetaB);
    if(P.size() > 0){
      printf("The PIM model is assumed.\n");
      printf("P_i = "); printVec(P);
      printf("Q_j = "); printVec(Q);
    }
    else{
      printf("The infinite-alleles model is assumed.\n");
    }
    cout << endl;
    // print probabilities
    printf("q_0                    = %+-1.10e\n",pzero);
    printf("q_1                    = %+-1.10e\n",pone);
    printf("q_2,E[g]               = %+-1.10e\n",ptwoEg);
    if(exactq2==true){
      printf("q_2(aVec+cA,bVec+cB,0) = %+-1.10e\n\n",ptwoRec);
    }
    else{
      printf("q_2(aVec+cA,bVec+cB,0) is not computed.  Use -e option to compute it.\n\n",ptwoRec);
    }
    
    printf("   rho    ASF for ordered configuration\n");
    for(iI=0; iI < rho.size(); iI++){
      asf = pzero+pone/rho[iI]+(ptwoEg+ptwoRec)/(rho[iI]*rho[iI]);
      if(asf >= 0){
	printf("%6.1f    %1.6e\n", rho[iI], asf);
      }
      else{
	printf("%6.1f    %1.6e  Rho is too small.  This answer is clearly not accurate.\n", rho[iI], asf);
      }
    }
  }
  else{
    for(iI=0; iI < rho.size(); iI++){
      printf("%1.6e  ", pzero+pone/rho[iI]+(ptwoEg+ptwoRec)/(rho[iI]*rho[iI]));
    }
    printf("\n");
  }

  return 0;
}
